class viterbi:
    _WORDTAG = 'WORDTAG'
    _UNIGRAM = '1-GRAM'
    _BIGRAM = '2-GRAM'
    _TRIGRAM = '3-GRAM'

    # public variable
    NO_VALUE = -1.0

    def __init__(self, get_rare, f_count):
        """
        get_rare: function to convert a word to rare marker
        f_count: count file such as gene.counts
        """
        self._tag_set_e = set(['*', 'STOP'])
        self._tag_set = set()
        self._word_set = set()
        self._e_dict = {}
        self._q_dict = {}
        self._pi_dict = {}
        self._get_rare = get_rare
        self._generate_dict(f_count)


    def _generate_dict(self, f_count):
        tag_dict = {}
        bigram_dict = {}
        with open(f_count, 'r') as fin:
            for l in fin:
                line = l.strip()
                if line:
                   fields = line.split(' ')
                   if fields[1] == self._WORDTAG:
                       key = fields[3], fields[2]
                       self._e_dict[key] = float(fields[0])
                       self._word_set.add(fields[3])
                   elif fields[1] == self._UNIGRAM:
                      self._tag_set.add(fields[2])
                      tag_dict[fields[2]] = float(fields[0])
                   elif fields[1] == self._BIGRAM:
                       key = fields[2], fields[3]
                       bigram_dict[key] = float(fields[0])
                   elif fields[1] == self._TRIGRAM:
                       key = fields[2], fields[3], fields[4]
                       self._q_dict[key] = float(fields[0])
        
        self._tag_set_e |= self._tag_set
        for a_key in self._e_dict.keys():
            self._e_dict[a_key] = self._e_dict[a_key] / tag_dict[a_key[1]]
        for a_key in self._q_dict.keys():
            yi_2, yi_1, yi = a_key
            bi_key = yi_2, yi_1
            self._q_dict[a_key] = self._q_dict[a_key] / bigram_dict[bi_key]


    def _get_e(self, word, tag):
        """
        Returns e(word, tag)
        If word is no in word_set, treat it as rare
        """
        if word not in self._word_set:
            word = self._get_rare(word)
        search_key = word, tag
        if search_key in self._e_dict.keys():
            return self._e_dict[search_key]
        else:
            return 0.0


    def _get_q(self, yi_2, yi_1, yi):
        if len(self._q_dict) == 0:
            return self.NO_VALUE
        search_key = yi_2, yi_1, yi
        if search_key in self._q_dict:
            return self._q_dict[search_key]
        else:
            return self.NO_VALUE


    def _get_pi(self, sentence, k, u, v):
        if k == 0:
            if u == "*" and v == "*":
                return 1.0, [u, v]
            else:
                return 0.0, [u, v]

        value_key = k, u, v
        # if we already have the result, just return it
        if value_key in self._pi_dict.keys():
            return self._pi_dict[value_key]

        word = sentence[k-1]
        e_p = self._get_e(word, v)
        max_pi = self.NO_VALUE
        max_tags = []
        if k == 2 or k == 1:
            max_pi, max_tags = self._get_pi(sentence, k - 1, "*", u)
            q = self._get_q("*", u, v)
            max_pi = max_pi * q * e_p
            max_tags = max_tags + [v]
        else:
            for w in self._tag_set:
                cur_pi, cur_tags = self._get_pi(sentence, k - 1, w, u)
                cur_pi = cur_pi * self._get_q(w, u, v) * e_p
                if cur_pi > max_pi:
                    max_pi = cur_pi
                    max_tags = cur_tags + [v]
        self._pi_dict[value_key] = [max_pi, max_tags]
        return max_pi, max_tags

    def get_tag_from_sentence(self, sentence):
        """
        Returns tag sequence as a list, given sentence.
        Input:
        sentence: list of words
        """
        self._pi_dict = {}
        max_pi = self.NO_VALUE
        max_tags = []
        for u in self._tag_set:
            for v in self._tag_set:
                cur_pi, cur_tags = self._get_pi(sentence, len(sentence), u, v)
                cur_pi = cur_pi * self._get_q(u, v, "STOP")
                if cur_pi > max_pi:
                    max_pi = cur_pi
                    max_tags = cur_tags + ["STOP"]
        return max_tags


    def get_tag_from_file(self, dev_file):
        sentence = []
        tags = []
        with open(dev_file, 'r') as fdev:
            for l in fdev:
                line = l.strip()
                if len(line) == 0:
                    # if a line is empty, a sentence just ended
                    tags = self.get_tag_from_sentence(sentence)
                    for i in xrange(0, len(sentence)):
                        print sentence[i], tags[i+2]	# tags include leading "*" "*"
                    print
                    sentence = []	# new sentence
                else:
                    sentence = sentence + [line]
